use crate::Error;
use axum::extract::Extension;
use axum::extract::Form;
use axum::response::Html;
use axum::response::Redirect;
use axum_sessions::extractors::ReadableSession;
use axum_sessions::extractors::WritableSession;
use pbkdf2::{
    password_hash::{rand_core::OsRng, PasswordHash, PasswordHasher, PasswordVerifier, SaltString},
    Pbkdf2,
};
use serde::Deserialize;
use sqlx::{Pool, Postgres};
use std::sync::Arc;
use tera::Context;
use tera::Tera;

#[derive(Clone, Debug, Deserialize)]
pub struct RawUser {
    username: String,
    password: String,
}

pub async fn create_user(
    Form(data): Form<RawUser>,
    Extension(pool): Extension<Arc<Pool<Postgres>>>,
) -> Result<Redirect, Error> {
    let handle = tokio::task::spawn_blocking(move || {
        let salt = SaltString::generate(&mut OsRng);
        let password_hash = Pbkdf2.hash_password(data.password.as_bytes(), &salt);

        password_hash.map(|p| p.to_string())
    });

    let hash = handle.await??;

    sqlx::query("INSERT INTO users (username, password_hash) VALUES ($1, $2)")
        .bind(data.username)
        .bind(hash)
        .execute(&*pool)
        .await?;

    Ok(Redirect::to("/login"))
}

pub async fn login_backend(
    Form(data): Form<RawUser>,
    Extension(pool): Extension<Arc<Pool<Postgres>>>,
    mut session: WritableSession,
) -> Result<Redirect, Error> {
    let (hash,): (String,) = sqlx::query_as("SELECT password_hash FROM users WHERE username=$1")
        .bind(&data.username)
        .fetch_one(&*pool)
        .await?;

    let handle = tokio::task::spawn_blocking(move || {
        let parsed_hash = PasswordHash::new(&hash)?;
        Pbkdf2
            .verify_password(data.password.as_bytes(), &parsed_hash)
            .map(|_| true)
    });

    handle.await??;

    // If we've gotten to this point, then the password was valid

    session.insert("logged_in_as", &data.username)?;

    Ok(Redirect::to("/"))
}

macro_rules! login_and_redirect {
    ($session:expr, $to:literal) => {
        if $session.get::<String>("logged_in_as").is_some() {
            return Err(::axum::response::Redirect::to($to).into());
        }
    };
}

pub async fn register_form(
    Extension(tera): Extension<Arc<Tera>>,
    session: ReadableSession,
) -> Result<Html<String>, Error> {
    login_and_redirect!(session, "/");
    let rendered = tera.render("users/register.html", &Context::new())?;

    Ok(Html(rendered))
}

pub async fn login_form(
    Extension(tera): Extension<Arc<Tera>>,
    session: ReadableSession,
) -> Result<Html<String>, Error> {
    login_and_redirect!(session, "/");
    let rendered = tera.render("users/login.html", &Context::new())?;

    Ok(Html(rendered))
}
